PARTICLE_OFFSET = Transform(Vec(0.15, 0.15, 0.6), QuatEuler(0.0, 180.0, -90.0))

PARTICLE_VELOCITY = {0, 0, -0.1}
PARTICLE_LIFETIME = 10

PARTICLE_DELAY = 3

EXPLODE_DELAY = 8

BLINDNESS_TIME = 2

function FloatProp(key, defualt)
    local path = "savegame.mod." .. key
    if not HasKey(path) then return defualt end
    return GetFloat(path)
end

--Explosion sphere properties
SHELL_COUNT = FloatProp("shellCount", 15)
SHELL_DELAY = FloatProp("shellDelay", 0.1)
SHELL_EXPLOSION_SIZE = FloatProp("shellGap", 3.5)

--Other explosion properties
EXPLOSION_FORCE = FloatProp("explosionForce", 500) * 1000
EXPLOSION_MASS_SCALING = false

--Laser
laser = 0

--Store which explosion shell is going to go off
explosionShell = 0

exploded = false

blindness = 0

dPositions = {}

function init()
    grenade = FindShape('grenade')

    defualtBloom = GetPostProcessingProperty("bloom")
    defualtGamma = GetPostProcessingProperty("gamma")

    -- DebugPrint(defualtBloom)
    -- DebugPrint(defualtGamma)
end

function tick(dt)
    -- DebugWatch("d", #dPositions)
    -- for i, v in ipairs(dPositions) do
    --     DebugCross(v,1,1,1,1)
    --     DebugWatch(i,v)
    -- end

    local time = GetTime()
    local grenadePos = GetShapeWorldTransform(grenade)

    ---------------------------Post processing---------------------------

    SetPostProcessing(blindness)
    blindness = blindness - (dt/BLINDNESS_TIME)
    blindness = clamp(blindness, 0, 1)

    --if it has already exploded, delete the laser and return so no further actions will be executed
    if exploded then
        if laser ~= 0 then
            Delete(laser)
            laser = 0
        end
        return
    end

    if time > EXPLODE_DELAY then

    ---------------------------Explosion force---------------------------

        QueryRequire("physical dynamic")
        QueryRejectBody(GetShapeBody(grenade));
        local bodies = QueryAabbBodies({-9999, -9999, -9999},{9999, 9999, 9999})

        for i, b in ipairs(bodies) do
            local p = GetBodyTransform(b).pos
            local m = GetBodyMass(b)
            local dir = VecSub(p, grenadePos.pos)

            --square distance function
            local d = (dir[1]*dir[1]) + (dir[2]*dir[2]) + (dir[3]*dir[3])

            local force = (1/d) * EXPLOSION_FORCE * dt
            if EXPLOSION_MASS_SCALING then force = force * math.sqrt(m) end

            ApplyBodyImpulse(b, grenadePos.pos, VecScale(dir, force))
            --DebugWatch("p", dir)
            --DebugWatch("p2", p)

            --DebugLine(p,VecAdd(p,VecScale(dir, 1/d)))
        end

    --------------------------------Laser--------------------------------
        if laser == 0 then
            laser = Spawn("MOD/laser.xml", Transform(grenadePos.pos, QuatEuler(0, 0, 0)), true)[1]
            blindness = 1
        end
        if laser ~= 0 then
            SetBodyTransform(laser, Transform(grenadePos.pos, QuatEuler(0, 0, 0)))
            SetBodyDynamic(laser,false)
        end
    end

    -------------------------------Explode-------------------------------
    --Script time at which the next shell will explode
    local nextExplodeTime = EXPLODE_DELAY + (explosionShell * SHELL_DELAY)

    if time > nextExplodeTime then
        local shellSize = explosionShell * SHELL_EXPLOSION_SIZE

        local shellCircumfrence = (shellSize * 2 * math.pi)/2
        local shellCount = math.floor(shellCircumfrence / SHELL_EXPLOSION_SIZE)

        for y=0, shellCount do
            local yRad = ((y/shellCount) * shellSize) - (math.pi/2)

            local ringHeight = math.sin(yRad) 
            local ringSize = math.cos(yRad)
            local ringCircumfence = ringSize * shellSize * 2 * math.pi

            local ringCount = math.floor(ringCircumfence / SHELL_EXPLOSION_SIZE)

            for x=0, ringCount do
                local xRad = (x/ringCount) * ringSize

                local spherePos = Vec(math.sin(x) * ringSize, ringHeight, math.cos(x) * ringSize)
                local localPos = VecScale(spherePos, shellSize)
                local pos = VecAdd(localPos, grenadePos.pos)

                --DebugWatch("shellsize", shellSize)
                --table.insert(dPositions, pos)
                MakeHole(pos,20,20,20)
            end
        end

        explosionShell = explosionShell + 1

        if explosionShell >= SHELL_COUNT then exploded = true end
    end

    ------------------------------Particles------------------------------
    if time > PARTICLE_DELAY then
        local particlePos = TransformToParentTransform(grenadePos, PARTICLE_OFFSET)

        ParticleType("smoke")
        ParticleTile(0)
        ParticleColor(1,0,0)
        ParticleAlpha(1.0, 0.0)
        ParticleRadius(0.1)
        ParticleCollide(1, 1, "constant", 0.1)
        ParticleGravity(1)

        --Find the particle velocity
        local vel = TransformToParentVec(particlePos, PARTICLE_VELOCITY)

        SpawnParticle(particlePos.pos, vel, PARTICLE_LIFETIME)
    end
end

function SetPostProcessing(fac)
    local newBloom = lerp(defualtBloom, 10, fac)
    local newGamma = lerp(defualtGamma, 20, fac)
    SetPostProcessingProperty("bloom", newBloom)
    SetPostProcessingProperty("gamma", newGamma)

    -- DebugWatch("gamma", newGamma)
    -- DebugWatch("bloom", newBloom)
    -- DebugWatch("fac", fac)
end

function lerp(a, b, t)
    return a + (t * (b-a))
end

function MapRange(t, f1, f2, t1, t2)
    t = t - f1
    t = t / (f2-f1)

    t = t * (t2-t1)
    t = t + t1

    return t
end

function clamp(val, min, max)
    return math.max(math.min(val, max), min)
end